
#include <glib.h>

#include "remote-login.h"
#include "defines.h"

#include "server.h"
#include "rdp-server.h"
#include "citrix-server.h"
#include "uccs-server.h"

enum {
	ERROR_SERVER_URI,
	ERROR_LOGIN
};

GList * config_file_citrix_servers = NULL;
GList * config_file_uccs_servers = NULL;
GList * config_file_rdp_servers = NULL;

/* Get the error domain for this module */
static GQuark
error_domain (void)
{
	static GQuark value = 0;
	if (value == 0) {
		value = g_quark_from_static_string("remote-login-service");
	}
	return value;
}

/* For a type of server list call the function to create the object
   from the keyfile group */
static void
config_file_server_list (GKeyFile * parsed, const gchar * group, Server * (*server_func) (GKeyFile *, const gchar *), GList ** list)
{
	if (g_key_file_has_key(parsed, CONFIG_MAIN_GROUP, group, NULL)) {
		gchar ** grouplist = g_key_file_get_string_list(parsed, CONFIG_MAIN_GROUP, group, NULL, NULL);
		int i = 0;
		gchar * groupsuffix;

		for (groupsuffix = grouplist[0], i = 0; groupsuffix != NULL; groupsuffix = grouplist[++i]) {
			Server * server = SERVER(server_func(parsed, groupsuffix));

			if (server == NULL) {
				/* Assume a relevant error is printed above */
				continue;
			}

			*list = g_list_append(*list, server);
		}

		g_strfreev(grouplist);
	}

	return;
}

static void
find_config_file (GKeyFile * parsed, const gchar * cmnd_line)
{
	GError * error = NULL;
	const gchar * file = DEFAULT_CONFIG_FILE;

	if (cmnd_line != NULL) {
		file = cmnd_line;
	}

	if (!g_key_file_load_from_file(parsed, file, G_KEY_FILE_NONE, &error)) {
		g_warning("Unable to parse config file '%s': %s", file, error->message);
		g_error_free(error);
		return;
	}

	if (!g_key_file_has_group(parsed, CONFIG_MAIN_GROUP)) {
		g_warning("Config file '%s' doesn't have group '" CONFIG_MAIN_GROUP "'", file);
		/* Probably should clear the keyfile, but there doesn't seem to be a way to do that */
		return;
	}

	config_file_server_list(parsed, CONFIG_UCCS_KEY, uccs_server_new_from_keyfile, &config_file_uccs_servers);
	config_file_server_list(parsed, CONFIG_CITRIX_KEY, citrix_server_new_from_keyfile, &config_file_citrix_servers);
	config_file_server_list(parsed, CONFIG_RDP_KEY, rdp_server_new_from_keyfile, &config_file_rdp_servers);

	return;
}

void
server_list_to_array (GVariantBuilder * builder, GList * items)
{
	GList * head = NULL;
	for (head = items; head != NULL; head = g_list_next(head)) {
		Server * server = SERVER(head->data);
		GVariant * variant = server_get_variant(server);
		g_variant_builder_add_value(builder, variant);
	}

	return;
}

static gboolean
handle_get_servers (RemoteLogin * rl, GDBusMethodInvocation * invocation, gpointer user_data)
{
	GVariant * array = NULL;

	if (g_list_length(config_file_citrix_servers) +
			g_list_length(config_file_uccs_servers) > 0) {
		GVariantBuilder builder;
		g_variant_builder_init(&builder, G_VARIANT_TYPE_ARRAY);

		server_list_to_array(&builder, config_file_uccs_servers);
		server_list_to_array(&builder, config_file_citrix_servers);
		server_list_to_array(&builder, config_file_rdp_servers);

		array = g_variant_builder_end(&builder);
	} else {
		array = g_variant_new_array(G_VARIANT_TYPE("sssa(svb{sv})"), NULL, 0);
	}

	g_dbus_method_invocation_return_value(invocation, g_variant_new_tuple(&array, 1));

	return TRUE;
}

/* Handle the GetServerForLogin DBus call */
static gboolean
handle_get_servers_login (RemoteLogin * rl, GDBusMethodInvocation * invocation, gpointer user_data)
{
	GVariant * params = g_dbus_method_invocation_get_parameters(invocation);
	const gchar * sender = g_dbus_method_invocation_get_sender(invocation);

	GVariant * child = NULL;
	const gchar * uri = NULL;

	child = g_variant_get_child_value(params, 0);
	uri = g_variant_get_string(child, NULL);
	g_variant_unref(child); /* fine as we know params is still ref'd */

	GList * lserver = NULL;
	Server * server = NULL;
	for (lserver = config_file_uccs_servers; lserver != NULL; lserver = g_list_next(lserver)) {
		server = SERVER(lserver->data);

		if (server == NULL) {
			continue;
		}

		if (g_strcmp0(server->uri, uri) == 0) {
			break;
		}
	}

	if (lserver == NULL) {
		/* Couldn't find something with that URI, we're done, thanks. */
		g_dbus_method_invocation_return_error(invocation,
		                                      error_domain(),
		                                      ERROR_SERVER_URI,
		                                      "Unable to find a server with the URI: '%s'",
		                                      uri);

		return TRUE;
	}

	/* Unlock the Server */
	const gchar * username = NULL;
	const gchar * password = NULL;
	gboolean allowcache = FALSE;

	child = g_variant_get_child_value(params, 1);
	username = g_variant_get_string(child, NULL);
	g_variant_unref(child); /* fine as we know params is still ref'd */

	child = g_variant_get_child_value(params, 2);
	password = g_variant_get_string(child, NULL);
	g_variant_unref(child); /* fine as we know params is still ref'd */

	child = g_variant_get_child_value(params, 3);
	allowcache = g_variant_get_boolean(child);
	g_variant_unref(child);

	/* Try to login, or else we error out */
	if (!uccs_server_unlock(UCCS_SERVER(server), sender, username, password)) {
		g_dbus_method_invocation_return_error(invocation,
		                                      error_domain(),
		                                      ERROR_LOGIN,
		                                      "Unable to login with the username: '%s'",
		                                      username);

		return TRUE;
	}

	GVariantBuilder builder;
	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);

	/* Only network, no caching yet */
	g_variant_builder_add_value(&builder, g_variant_new_string("network"));

	/* Get the array of servers */
	GVariant * array = uccs_server_get_servers(UCCS_SERVER(server), sender);
	g_variant_builder_add_value(&builder, array);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));
	return TRUE;
}

static void
name_lost (GDBusConnection * connection, const gchar * name, gpointer user_data)
{
	GMainLoop * mainloop = (GMainLoop *)user_data;

	g_error("Unable to get name '%s'", name);
	g_main_loop_quit(mainloop);

	return;
}

static gchar * cmnd_line_config = NULL;

static GOptionEntry general_options[] = {
	{"config-file",  'c',  0,  G_OPTION_ARG_FILENAME,  &cmnd_line_config, "Configuration file for the remote login service.  Defaults to '" DEFAULT_CONFIG_FILE "'.", "key_file"},
	{NULL}
};

int
main (int argc, char * argv[])
{
	GError * error = NULL;

	/* Init the GTypes */
	g_type_init();
	GKeyFile * config = g_key_file_new();
	GMainLoop * mainloop = g_main_loop_new(NULL, FALSE);

	/* Handle command line parameters */
	GOptionContext * context;
	context = g_option_context_new("- Determine the remote servers that can be logged into");
	g_option_context_add_main_entries(context, general_options, "remote-login-service");

	if (!g_option_context_parse(context, &argc, &argv, &error)) {
		g_print("option parsing failed: %s\n", error->message);
		g_error_free(error);
		return 1;
	}

	/* Parse config file */
	find_config_file(config, cmnd_line_config);

	/* Start up D' Bus */
	GDBusConnection * session_bus = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL /* cancel */, &error);
	if (error != NULL) {
		g_error("Unable to get session bus: %s", error->message);
		g_error_free(error);
		return -1;
	}

	RemoteLogin * skel = remote_login_skeleton_new();
	g_dbus_interface_skeleton_export(G_DBUS_INTERFACE_SKELETON(skel),
	                                 session_bus,
	                                 "/com/canonical/RemoteLogin",
	                                 NULL);
	g_signal_connect(skel, "handle-get-servers", G_CALLBACK(handle_get_servers), NULL);
	g_signal_connect(skel, "handle-get-servers-for-login", G_CALLBACK(handle_get_servers_login), NULL);

	g_bus_own_name_on_connection(session_bus,
	                             "com.canonical.RemoteLogin",
	                             G_BUS_NAME_OWNER_FLAGS_NONE,
	                             NULL, /* aquired handler */
	                             name_lost,
	                             mainloop,
	                             NULL); /* mainloop free */

	/* Loop forever */
	g_main_loop_run(mainloop);

	g_main_loop_unref(mainloop);
	g_object_unref(config);

	g_free(cmnd_line_config);

	return 0;
}
