#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <json-glib/json-glib.h>

#include "uccs-server.h"
#include "defines.h"

#include "rdp-server.h"
#include "citrix-server.h"

static void uccs_server_class_init (UccsServerClass *klass);
static void uccs_server_init       (UccsServer *self);
static void uccs_server_dispose    (GObject *object);
static void uccs_server_finalize   (GObject *object);
static GVariant *  get_properties        (Server * server);

G_DEFINE_TYPE (UccsServer, uccs_server, SERVER_TYPE);

static void
uccs_server_class_init (UccsServerClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->dispose = uccs_server_dispose;
	object_class->finalize = uccs_server_finalize;

	ServerClass * server_class = SERVER_CLASS(klass);

	server_class->get_properties = get_properties;
	/* UCCS can't have applications */
	server_class->get_applications = NULL;

	return;
}

static void
uccs_server_init (UccsServer *self)
{
	self->exec = g_strdup(UCCS_QUERY_TOOL);

	self->username = NULL;
	self->password = NULL;

	self->lovers = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);

	self->subservers = NULL;

	return;
}

struct _hash_helper {
	GVariant * params;
	GDBusConnection * session;
};

/* GHashTable foreach item */
static gboolean
clear_hash_helper (gpointer key, gpointer value, gpointer user_data)
{
	struct _hash_helper * helper = (struct _hash_helper *)user_data;
	GError * error = NULL;

	g_dbus_connection_emit_signal(helper->session,
	                              (const gchar *)key, /* dest */
	                              "/com/canonical/RemoteLogin", /* object path */
	                              "com.canonical.RemoteLogin", /* interface name */
	                              "com.canonical.RemoteLogin.LoginChanged", /* signal name */
	                              helper->params, /* params */
	                              &error);

	if (error != NULL) {
		g_warning("Unable to signal UCCS server shutdown: %s", error->message);
		g_error_free(error);
	}

	return;
}

/* Clear the hash table by going through it and signaling */
static void
clear_hash (UccsServer * server)
{
	if (g_hash_table_size(server->lovers) == 0) {
		return;
	}

	g_return_if_fail(server->parent.uri != NULL);
	g_return_if_fail(server->username != NULL);

	GDBusConnection * session = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL, NULL); /* Shouldn't block, we should have it */
	GVariant * param = g_variant_new("(ss)", server->parent.uri, server->username); /* params */
	g_variant_ref_sink(param);

	struct _hash_helper helper;
	helper.params = param;
	helper.session = session;

	g_hash_table_foreach_remove(server->lovers, clear_hash_helper, &helper);

	g_object_unref(session);
	g_variant_unref(param);

	return;
}

/* Clean up references */
static void
uccs_server_dispose (GObject *object)
{
	UccsServer * self = UCCS_SERVER(object);

	if (self->lovers != NULL) {
		clear_hash(self);
	}

	g_list_free_full(self->subservers, g_object_unref);
	self->subservers = NULL; /* Ironically the free function is the only GList
	                      function that doesn't return a new pointer by itself */

	G_OBJECT_CLASS (uccs_server_parent_class)->dispose (object);
	return;
}

/* Clean up memory */
static void
uccs_server_finalize (GObject *object)
{
	UccsServer * self = UCCS_SERVER(object);

	g_free(self->exec); self->exec = NULL;
	g_free(self->username); self->username = NULL;
	g_free(self->password); self->password = NULL;

	if (self->lovers != NULL) {
		g_hash_table_unref(self->lovers);
		self->lovers = NULL;
	}

	G_OBJECT_CLASS (uccs_server_parent_class)->finalize (object);
	return;
}

static GVariant *
get_properties (Server * server)
{
	GVariantBuilder propbuilder;
	g_variant_builder_init(&propbuilder, G_VARIANT_TYPE_ARRAY);

	GVariantBuilder namebuilder;
	g_variant_builder_init(&namebuilder, G_VARIANT_TYPE_TUPLE);
	g_variant_builder_add_value(&namebuilder, g_variant_new_string("email"));
	g_variant_builder_add_value(&namebuilder, g_variant_new_boolean(TRUE));
	g_variant_builder_add_value(&namebuilder, g_variant_new_variant(g_variant_new_string("")));
	g_variant_builder_add_value(&namebuilder, g_variant_new((const gchar *)G_VARIANT_TYPE_VARDICT));
	g_variant_builder_add_value(&propbuilder, g_variant_builder_end(&namebuilder));

	GVariantBuilder passbuilder;
	g_variant_builder_init(&passbuilder, G_VARIANT_TYPE_TUPLE);
	g_variant_builder_add_value(&passbuilder, g_variant_new_string("password"));
	g_variant_builder_add_value(&passbuilder, g_variant_new_boolean(TRUE));
	g_variant_builder_add_value(&passbuilder, g_variant_new_variant(g_variant_new_string("")));
	g_variant_builder_add_value(&passbuilder, g_variant_new((const gchar *)G_VARIANT_TYPE_VARDICT));
	g_variant_builder_add_value(&propbuilder, g_variant_builder_end(&passbuilder));

	return g_variant_builder_end(&propbuilder);
}

Server *
uccs_server_new_from_keyfile (GKeyFile * keyfile, const gchar * name)
{
	g_return_val_if_fail(keyfile != NULL, NULL); /* NOTE: No way to check if that's really a keyfile :-( */
	g_return_val_if_fail(name != NULL, NULL);

	gchar * groupname = g_strdup_printf("%s %s", CONFIG_UCCS_PREFIX, name);

	if (!g_key_file_has_group(keyfile, groupname)) {
		g_warning("Server '%s' specified but group '%s' was not found", name, groupname);
		g_free(groupname);
		return NULL;
	}

	UccsServer * server = g_object_new(UCCS_SERVER_TYPE, NULL);

	if (g_key_file_has_key(keyfile, groupname, "Name", NULL)) {
		server->parent.name = g_key_file_get_string(keyfile, groupname, CONFIG_SERVER_NAME, NULL);
	}

	if (g_key_file_has_key(keyfile, groupname, "URI", NULL)) {
		server->parent.uri = g_key_file_get_string(keyfile, groupname, CONFIG_SERVER_URI, NULL);
	}

	if (g_key_file_has_key(keyfile, groupname, CONFIG_UCCS_EXEC, NULL)) {
		g_free(server->exec);
		server->exec = g_key_file_get_string(keyfile, groupname, CONFIG_UCCS_EXEC, NULL);
	}

	g_free(groupname);
	return SERVER(server);
}

/* Call the external utility and get the JSON data for the servers of this
   UCCS server */
static gchar *
get_json (UccsServer * server)
{
	gint exit_status = 0;
	gchar * output = NULL;
	gchar * eoutput = NULL;
	GError * error = NULL;

	gchar * cmdline = g_strdup_printf(HELPER_SCRIPT " %s %s %s", server->username, server->password, server->exec);

	g_spawn_command_line_sync(cmdline, &output, &eoutput, &exit_status, &error);

	if (eoutput != NULL && eoutput[0] != '\0') {
		g_debug("Script output stderr: '%s'", eoutput);
	}
	g_free(eoutput);

	if (error != NULL) {
		g_warning("Unable to execute script for JSON data: %s", error->message);
		g_error_free(error);
		return NULL;
	}

	return output;
}

/* Look at the array of RLS data and build a server for each entry
   in the array */
static gboolean
parse_rds_array (UccsServer * server, JsonArray * array)
{
	int i;
	for (i = 0; i < json_array_get_length(array); i++) {
		JsonNode * node = json_array_get_element(array, i);

		if (JSON_NODE_TYPE(node) != JSON_NODE_OBJECT) {
			continue;
		}

		JsonObject * object = json_node_get_object(node);
		if (!json_object_has_member(object, "Protocol")) {
			continue;
		}

		JsonNode * proto_node = json_object_get_member(object, "Protocol");
		if (JSON_NODE_TYPE(proto_node) != JSON_NODE_VALUE) {
			continue;
		}
		if (json_node_get_value_type(proto_node) != G_TYPE_STRING) {
			continue;
		}

		const gchar * proto = json_node_get_string(proto_node);
		Server * newserver = NULL;

		if (g_strcmp0(proto, "ICA") == 0) {
			newserver = citrix_server_new_from_json(object);
		}
		else if (g_strcmp0(proto, "freerdp") == 0) {
			newserver = rdp_server_new_from_json(object);
		}

		if (newserver == NULL) continue;

		server->subservers = g_list_append(server->subservers, newserver);
	}

	return TRUE;
}

/* Parse the JSON content and allocate servers based on that */
static gboolean
parse_json (UccsServer * server, const gchar * json)
{
	if (json == NULL) return; /* Shouldn't happen, but let's just handle it */

	gboolean passed = TRUE;
	JsonParser * parser = json_parser_new();
	GError * error = NULL;

	if (!json_parser_load_from_data(parser, json, -1, &error)) {
		g_warning("Unable to parse JSON data: %s", error->message);
		g_error_free(error);
		error = NULL;
		passed = FALSE;
	}

	/* Make sure we have a sane root node */
	JsonNode * root_node = NULL;
	if (passed) {
		root_node = json_parser_get_root(parser);
	}
	if (root_node != NULL && JSON_NODE_TYPE(root_node) != JSON_NODE_OBJECT) {
		g_warning("Root node of JSON data is not an object.  It is: %s", json_node_type_name(root_node));
		passed = FALSE;
	}

	/* Take our object and see if it has the property that we need */
	JsonObject * root_object = NULL;
	if (passed) {
		root_object = json_node_get_object(root_node);
	}
	if (root_object != NULL && json_object_has_member(root_object, "RemoteDesktopServers")) {
		/* This shows that we have some.  It's okay if there aren't any.  Seems
		   a bit silly, but we're not bitching too much. */
		JsonNode * rds_node = json_object_get_member(root_object, "RemoteDesktopServers");
		if (JSON_NODE_TYPE(rds_node) == JSON_NODE_ARRAY) {
			JsonArray * rds_array = json_node_get_array(rds_node);
			passed = parse_rds_array(server, rds_array);
		} else {
			/* Okay we're a little bit angrier about this one */
			g_warning("Malformed 'RemoteDesktopServer' entry.  Not an array but a: %s", json_node_type_name(rds_node));
			passed = FALSE;
		}
	} else {
		g_debug("No 'RemoteDesktopServers' found");
	}

	g_object_unref(parser);
	return passed;
}

/**
 * uccs_server_unlock:
 * @server: The server to unlock
 * @address: DBus address of the person unlocking us
 * @username: Username for the UCCS
 * @password: (allow-none) Password to use
 *
 * Unlocks the UCCS server making servers available either from the
 * cache or from the network.
 *
 * Return value: Returns true if the creditials worked
 */
gboolean
uccs_server_unlock (UccsServer * server, const gchar * address, const gchar * username, const gchar * password)
{
	g_return_val_if_fail(IS_UCCS_SERVER(server), FALSE);
	g_return_val_if_fail(username != NULL, FALSE);
	g_return_val_if_fail(address != NULL, FALSE);

	/* Check the current values we have, they might be NULL, which in
	   that case they won't match */
	if (g_strcmp0(username, server->username) == 0 &&
			g_strcmp0(password, server->password) == 0) {
		g_hash_table_insert(server->lovers, g_strdup(address), GINT_TO_POINTER(TRUE));
		return TRUE;
	}

	if (server->username != NULL && (server->password != NULL || password == NULL)) {
		/* We've got stored values, but those weren't them */
		return FALSE;
	}

	g_return_val_if_fail(server->exec != NULL, FALSE); /* Shouldn't happen, but I'd feel safer if we checked */

	/* We're changing the username and password, if there were other
	   people who had it, they need to know we're different now */
	clear_hash(server);
	server->username = g_strdup(username);
	server->password = g_strdup(password);

	gchar * json = get_json(server);

	if (json == NULL) {
		g_free(server->username);
		server->username = NULL;
		g_free(server->password);
		server->password = NULL;
		return FALSE;
	}

	gboolean parser = parse_json(server, json);
	g_free(json);

	if (parser) {
		g_hash_table_insert(server->lovers, g_strdup(address), GINT_TO_POINTER(TRUE));
	}

	return parser;
}

/* A little quickie function to handle the null server array */
inline static GVariant *
null_server_array (void)
{
	return g_variant_new_array(G_VARIANT_TYPE("(sssba(sbva{sv})a(si))"),
		NULL, 0);
}

/**
 * uccs_server_get_servers:
 * @server: Server to get our list from
 * @address: Who's asking
 *
 * Will get a valid variant with servers.  If the asker hasn't unlocked us
 * then the list will always be empty.
 *
 * Return value: A variant array
 */
GVariant *
uccs_server_get_servers (UccsServer * server, const gchar * address)
{
	g_return_val_if_fail(IS_UCCS_SERVER(server), null_server_array());
	g_return_val_if_fail(address != NULL, null_server_array());

	if (!GPOINTER_TO_INT(g_hash_table_lookup(server->lovers, address))) {
		g_warning("Address '%s' is not authorized", address);
		return null_server_array();
	}

	if (g_list_length(server->subservers) == 0) {
		return null_server_array();
	}

	GVariantBuilder array;
	g_variant_builder_init(&array, G_VARIANT_TYPE_ARRAY);
	GList * lserver;
	for (lserver = server->subservers; lserver != NULL; lserver = g_list_next(lserver)) {
		Server * serv = SERVER(lserver->data);
		g_variant_builder_add_value(&array, server_get_variant(serv));
	}

	return g_variant_builder_end(&array);
}
