
#include <glib.h>
#include <gio/gio.h>
#include <libdbustest/dbus-test.h>

typedef struct _slmock_table_t slmock_table_t;
typedef struct _slmock_server_t slmock_server_t;

struct _slmock_table_t {
	const gchar * username;
	const gchar * password;
	slmock_server_t * servers;
};

struct _slmock_server_t {
	const gchar * name;
	const gchar * uri;
	const gchar * type;
	/* TODO: Add more as we get more working */
};

slmock_server_t citrix_server_table[] = {
	{"Citrix USA", "107.21.17.35", "ica"},
	{"Citrix 2",   "107.21.17.35", "ica"},
	{"Citrix 3",   "107.21.17.35", "ica"},
	{NULL, NULL, NULL}
};

slmock_server_t freerdp_server_table[] = {
	{"FreeRDP US",    "23.21.151.133",  "rdp"},
	{"FreeRDP Asia",  "46.137.222.181", "rdp"},
	{"FreeRDP UK",    "46.137.189.194", "rdp"},
	{NULL, NULL, NULL}
};

slmock_server_t big_server_table[] = {
	{"XenServer",    "107.21.17.35",    "ica"},
	{"Citrix2",      "http://1.2.3.4",  "ica"},
	{"Accenture",    "10.21.17.35",     "rdp"},
	{"Accenture 2",  "https://4.5.6.7", "rdp"},
	{NULL, NULL, NULL}
};

slmock_table_t slmock_table[] = {
	{"c", "c", citrix_server_table},
	{"f", "f", freerdp_server_table},
	{"b", "b", big_server_table}
};

static gboolean
find_server (GVariant * varray, slmock_server_t * server)
{
	int ichild;
	for (ichild = 0; ichild < g_variant_n_children(varray); ichild++) {
		GVariant * child = g_variant_get_child_value(varray, ichild);
		gboolean test = FALSE;

		GVariant * type = g_variant_get_child_value(child, 0);
		test = g_strcmp0(g_variant_get_string(type, NULL), server->type) == 0;
		g_variant_unref(type);
		if (!test) {
			g_variant_unref(child);
			continue;
		}

		GVariant * name = g_variant_get_child_value(child, 1);
		test = g_strcmp0(g_variant_get_string(name, NULL), server->name) == 0;
		g_variant_unref(name);
		if (!test) {
			g_variant_unref(child);
			continue;
		}

		GVariant * uri = g_variant_get_child_value(child, 2);
		test = g_strcmp0(g_variant_get_string(uri, NULL), server->uri) == 0;
		g_variant_unref(uri);
		if (!test) {
			g_variant_unref(child);
			continue;
		}

		g_variant_unref(child);
		return TRUE;
	}

	return FALSE;
}

static gboolean
slmock_check_login(GDBusConnection * session, slmock_table_t * slmockdata)
{
	GVariant * retval = g_dbus_connection_call_sync(session,
	                                                "com.canonical.RemoteLogin",
	                                                "/com/canonical/RemoteLogin",
	                                                "com.canonical.RemoteLogin",
	                                                "GetServersForLogin",
	                                                g_variant_new("(sssb)",
	                                                              "https://slmock.com/",
	                                                              slmockdata->username,
	                                                              slmockdata->password,
	                                                              TRUE), /* params */
	                                                G_VARIANT_TYPE("(sa(sssba(sbva{sv})a(si)))"), /* ret type */
	                                                G_DBUS_CALL_FLAGS_NONE,
	                                                -1,
	                                                NULL,
	                                                NULL);

	g_assert(retval != NULL);
	g_assert(g_variant_n_children(retval) == 2);

	GVariant * data = g_variant_get_child_value(retval, 0);
	g_assert(g_strcmp0(g_variant_get_string(data, NULL), "network") == 0);
	g_variant_unref(data); data = NULL;

	GVariant * array = g_variant_get_child_value(retval, 1);
	int i;
	for (i = 0; slmockdata->servers[i].name != NULL; i++) {
		g_assert(find_server(array, &slmockdata->servers[i]));
	}

	g_variant_unref(retval);

	return TRUE;
}

static void
test_getservers_slmock (gconstpointer data)
{
	DbusTestService * service = dbus_test_service_new(NULL);

	/* RLS */
	DbusTestProcess * rls = dbus_test_process_new(REMOTE_LOGIN_SERVICE);
	dbus_test_process_append_param(rls, "--config-file=" SLMOCK_CONFIG_FILE);
	dbus_test_service_add_task(service, DBUS_TEST_TASK(rls));

	/* Dummy */
	DbusTestTask * dummy = dbus_test_task_new();
	dbus_test_task_set_wait_for(dummy, "com.canonical.RemoteLogin");
	dbus_test_service_add_task(service, dummy);

	/* Get RLS up and running and us on that bus */
	dbus_test_service_start_tasks(service);

	GDBusConnection * session = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL, NULL);
	g_dbus_connection_set_exit_on_close(session, FALSE);

	g_assert(slmock_check_login(session, (slmock_table_t *)data));

	g_object_unref(session);
	g_object_unref(rls);
	g_object_unref(service);

	return;
}

static void
test_getservers_uccs (void)
{
	DbusTestService * service = dbus_test_service_new(NULL);

	/* RLS */
	DbusTestProcess * rls = dbus_test_process_new(REMOTE_LOGIN_SERVICE);
	dbus_test_process_append_param(rls, "--config-file=" UCCS_CONFIG_FILE);
	dbus_test_service_add_task(service, DBUS_TEST_TASK(rls));

	/* Dummy */
	DbusTestTask * dummy = dbus_test_task_new();
	dbus_test_task_set_wait_for(dummy, "com.canonical.RemoteLogin");
	dbus_test_service_add_task(service, dummy);

	/* Get RLS up and running and us on that bus */
	dbus_test_service_start_tasks(service);

	GDBusConnection * session = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL, NULL);
	g_dbus_connection_set_exit_on_close(session, FALSE);

	GVariant * retval = g_dbus_connection_call_sync(session,
	                                                "com.canonical.RemoteLogin",
	                                                "/com/canonical/RemoteLogin",
	                                                "com.canonical.RemoteLogin",
	                                                "GetServers",
	                                                NULL, /* params */
	                                                G_VARIANT_TYPE("(a(sssba(sbva{sv})a(si)))"), /* ret type */
	                                                G_DBUS_CALL_FLAGS_NONE,
	                                                -1,
	                                                NULL,
	                                                NULL);

	g_assert(retval != NULL);
	g_assert(g_variant_n_children(retval) == 1);

	GVariant * array = g_variant_get_child_value(retval, 0);
	g_assert(g_variant_n_children(array) == 1);
	g_assert(g_variant_is_of_type(array, G_VARIANT_TYPE_ARRAY));

	GVariant * tuple = g_variant_get_child_value(array, 0);

	if (TRUE) { /* type check */
		GVariant * type = g_variant_get_child_value(tuple, 0);
		g_assert(g_variant_is_of_type(type, G_VARIANT_TYPE_STRING));
		g_assert(g_strcmp0(g_variant_get_string(type, NULL), "uccs") == 0);
		g_variant_unref(type);
	}

	if (TRUE) { /* name check */
		GVariant * name = g_variant_get_child_value(tuple, 1);
		g_assert(g_variant_is_of_type(name, G_VARIANT_TYPE_STRING));
		g_assert(g_strcmp0(g_variant_get_string(name, NULL), "Test Server Name") == 0);
		g_variant_unref(name);
	}

	if (TRUE) { /* uri check */
		GVariant * uri = g_variant_get_child_value(tuple, 2);
		g_assert(g_variant_is_of_type(uri, G_VARIANT_TYPE_STRING));
		g_assert(g_strcmp0(g_variant_get_string(uri, NULL), "https://uccs.test.mycompany.com/") == 0);
		g_variant_unref(uri);
	}

	g_variant_unref(tuple);
	g_variant_unref(array);
	g_variant_unref(retval);

	g_object_unref(session);
	g_object_unref(rls);
	g_object_unref(service);

	return;
}


/* Build the test suite */
static void
test_dbus_suite (void)
{
	g_test_add_func ("/dbus/interface/GetServers/UCCS",   test_getservers_uccs);
	g_test_add_data_func ("/dbus/interface/GetServers/SLMock/citrix",  &slmock_table[0], test_getservers_slmock);
	g_test_add_data_func ("/dbus/interface/GetServers/SLMock/freerdp", &slmock_table[1], test_getservers_slmock);
	g_test_add_data_func ("/dbus/interface/GetServers/SLMock/big",     &slmock_table[2], test_getservers_slmock);

	return;
}

gint
main (gint argc, gchar * argv[])
{
	g_type_init();
	g_test_init(&argc, &argv, NULL);

	/* Test suites */
	test_dbus_suite();

	return g_test_run ();
}
